namespace TP
{
	/// <summary>
	/*	 This namespace provides the interfaces that define this project
		 We also provide files implementing some limited functionality for
		 these interfaces. You are welcome to write them yourself if you
		 prefer. */		 
	/// </summary>
 	[System.Serializable()]
	public class Transaction
	{
		/// <summary>
		/*   Transaction Identifier */
		/// </summary>
		public System.Guid Id;

		/// <summary>
		/*   Constructor */
		/// </summary> 
		public Transaction()
		{
			this.Id = System.Guid.NewGuid();
		}

		/* Override Equals and GetHashCode so that hashtables hash on the TX Guid */
		public override bool Equals(object o)
		{
			Transaction tx = o as Transaction;
			if(tx==null) return false;

			return tx.Id.Equals(this.Id);
		}

		public override int GetHashCode()
		{
			return this.Id.GetHashCode();
		}
	}

	/// <summary>
	/*   Customer class */
	/// </summary>
	[System.Serializable()]
	public class Customer
	{
		/* Customer Identifier */
		public System.Guid Id;

		/// <summary>
		/// Constructor. Generates a new Id for the customer
		/// </summary>
		public Customer()
		{
			this.Id = new System.Guid();
		}

		public override int GetHashCode()
		{
			// Debug: 
			// Console.WriteLine(this+" GetHashCode="+this.Name.GetHashCode());

			return this.Id.GetHashCode ();
		}

		public override bool Equals(object obj)
		{
			// Debug: 
			// Console.WriteLine(this+" Equals "+obj+" = "+this.Name.Equals());

			TP.Customer cust = obj as TP.Customer;
			if(cust!=null) 
				return this.Id.Equals(cust.Id);

			return false;
		}
	}

	/// <summary>
	/*   Reservable Item class */
	/// </summary>
	[System.Serializable()]
	public class Item
	{
		public string Name;

		public Item(string name)
		{
			this.Name = name;
		}

		public Item(int name)
		{
			this.Name = name.ToString();
		}

		public override int GetHashCode()
		{
			// Debug:
			// Console.WriteLine(this+" GetHashCode="+this.Name.GetHashCode());
			
			return this.Name.GetHashCode ();
		}

		public override bool Equals(object obj)
		{
			// Debug:
			// Console.WriteLine(this+" Equals "+obj+" = "+this.Name.Equals());
			
			TP.Item item = obj as TP.Item;
			if(item!=null)
			{
				return this.Name.Equals(item.Name);
			}
			
			return false;
		}
	}

	/// <summary>
	/*   Resource interface */
	/// </summary>
	public interface RS{}

	/// <summary>
	/// Lock manager interface
	/// </summary>
	public interface LM
	{
		/// <param name="context"></param>
		/// <param name="resource"></param>
		void LockForRead(Transaction context, RS resource);

		/// <param name="context"></param>
		/// <param name="resource"></param>
		void LockForWrite(Transaction context, RS resource);
		
		/// <param name="context"></param>
		void UnlockAll(Transaction context);
	}

	/// <summary>
	/*   Transaction manager interface */
	/// </summary>
	public interface TM
	{
		Transaction Start();

		/// <param name="context"></param>
		void Commit(Transaction context);

		/// <param name="context"></param>
		void Abort(Transaction context);
		
		/// <summary>
		/*   Enlist the RM as a member of this transaction */
		/// </summary>
		/// <param name="context"></param>
		void Enlist(Transaction context, RM enlistingRM);
	}

	/// <summary>
	/*   Workflow controller interface */
	/// </summary>
	public interface WC
	{
		/// <param name="c"></param>
		/// <param name="flights"></param>
		/// <param name="location"></param>
		/// <param name="car"></param>
		/// <param name="room"></param>
		/// <returns>success</returns>
		bool ReserveItinerary(TP.Customer c, int[] flights, string location, bool car, bool room);

		/* This implies calling the Add method of the Resource Manager that handles flights.
		   Item here is usually the location name */
		bool AddFlight(Transaction context, Item i, int count, int price );
		bool DeleteFlight(Transaction context, Item i);
		bool AddRooms(Transaction context, Item i, int count, int price );
		bool DeleteRooms(Transaction context, Item i);
		bool AddCars(Transaction context, Item i, int count, int price );
		bool DeleteCars(Transaction context, Item i);
		int  QueryFlight(Transaction context, Item i);
		int  QueryFlightPrice(Transaction context, Item i);
		int  QueryRooms(Transaction context, Item i);
		int  QueryRoomsPrice(Transaction context, Item i);
		int  QueryCars(Transaction context, Item i);
		int  QueryCarsPrice(Transaction context, Item i);
		string  QueryCustomerInfo(TP.Transaction context,TP.Customer c);
		int  QueryCustomer(TP.Transaction context,TP.Customer c);
	}

	/// <summary>
	/*   Resource Manager Interface */
	/// </summary>
	public interface RM
	{
		/// <param name="context"></param>
 		void Enlist(Transaction context);

		/// <param name="context"></param>
		void Commit(Transaction context);

		/// <param name="context"></param>
		void Abort(Transaction context);
	
		/// <summary>
		/*   Add "count" items 
		     This method is equivalent to the addXXX in the Java Interface */
		/// </summary>
		/// <param name="context"></param>
		/// <param name="i"></param>
		/// <param name="count"></param>
		/// <param name="price"></param>
		/// <returns>success</returns>
		bool Add(Transaction context, Item i, int count, int price);

		bool Delete(Transaction context, Item i);
	
		/// <summary>
		/*   Query: equivalent to QueryCars, QueryFlights, QueryRooms in the Java interface */
		/// </summary>
		/// <param name="context"></param>
		/// <param name="i"></param>
		/// <returns></returns>
 		int Query(Transaction context, Item i);

		/// <param name="context"></param>
		/// <param name="i"></param>
		/// <returns></returns>
		int QueryPrice(Transaction context, Item i);

		string QueryCustomerInfo(Transaction context, Customer c);

		/// <param name="context"></param>
		/// <param name="c"></param>
		/// <returns></returns>
		int QueryCustomer(Transaction context, Customer c);

		/// <param name="context"></param>
		/// <param name="c"></param>
		/// <param name="i"></param>
		/// <returns>sucess</returns>
		bool Reserve(Transaction context, Customer c, Item i); 

		/// <param name="context"></param>
		/// <param name="c"></param>
		void DropReservations(Transaction context, Customer c);
		
		/// <summary>
		/*   Shutdown should gracefully clean up its files, so when it
			 restarts, it should not need to recover the state  */
		///  </summary>
		void Shutdown();

		/// <summary>
		/*   Exit (simulate a failure) after a specified number of disk writes.
		     Support for this method requires a wrapper around the system's 
		     write system call that decrements the counter set by this method.
		     This counter should be set to zero by default, which makes the wrapper
		     doing nothing.  If the counter is non-zero, the wrapper should decrement								     
		     it, see if it is zero, and if so call exit().
		     This method is not part of a transaction. It is intended to simulate an
			 RM failure. */
		/// </summary>
		/// <param name="diskWritesToWait"></param>
		void SelfDestruct(int diskWritesToWait);
	}
}
